from promptops.loading.base import LoadingBase


class Pong(LoadingBase):
    _DOTS = "⠁⠂⠄⡀⠈⠐⠠⢀"
    _PADDLES = "╵╷"

    def __init__(self, width=5):
        super().__init__()
        self._use_left_dots = True
        self._width = width
        self._height = 4
        self._x = 0
        self._y = 0
        self._dx = 1
        self._dy = 1
        self._paddle_left = 0
        self._paddle_right = 1

    def _draw(self):
        s = self._PADDLES[self._paddle_left] + " " * self._x + self._DOTS[self._y + (0 if self._use_left_dots else self._height)] + " " * (self._width - 1 - self._x) + self._PADDLES[self._paddle_right]
        self._stream.write(s + "\n")
        self._written_lines = 2

        self._x += self._dx
        self._y += self._dy
        if self._x < 0 or self._x >= self._width:
            self._dx *= -1
            self._x = max(0, min(self._x, self._width - 1))
            self._use_left_dots = not self._use_left_dots
        if self._y < 0 or self._y >= self._height:
            self._dy *= -1
            self._y = max(0, min(self._y, self._height - 1))
        if self._dx > 0:
            self._paddle_right = self._y // (self._height // 2)
        else:
            self._paddle_left = self._y // (self._height // 2)



from dataclasses import dataclass

import requests

from promptops import settings
from promptops import trace
from promptops import user


@dataclass
class Result:
    script: str
    explanation: str = None
    lang: str = "shell"
    origin: str = "promptops"

    @staticmethod
    def from_dict(data: dict) -> 'Result':
        return Result(
            script=data["script"],
            explanation=data.get("explanation"),
            lang=data.get("lang", "shell"),
            origin=data.get("origin", "promptops"),
        )


def query(*, q: str) -> list[Result]:
    response = requests.post(settings.endpoint + "/query", json={
        "query": q,
        "explanation": settings.request_explanation,
        "model": settings.model,
        "trace_id": trace.trace_id,
    }, headers={
        "user-agent": f"promptops-cli; user_id={user.user_id}",
    })
    if response.status_code != 200:
        raise Exception(f"there was problem with the response, status: {response.status_code}, text: {response.text}")
    data = response.json()
    return [Result.from_dict(entry) for entry in data["suggestions"]]



import logging
import time
import random

import requests
import colorama

from promptops import corrections
from .dtos import Result
import threading
from promptops.loading import Simple
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML, to_formatted_text
from typing import Callable
from promptops import settings
from promptops.user import user_id
from promptops.trace import trace_id
from . import messages
from promptops import scrub_secrets


def get_explanation(
    *,
    result: Result,
    prev_questions: list[str],
    prev_results: list[list[str]],
    corrected_results: list[corrections.QATuple],
    history_context: list[str],
    similar_history: list[str],
):
    response = requests.post(
        settings.endpoint + "/explain",
        json={
            "script": result.script,
            "lang": result.lang,
            "questions": prev_questions,
            "results": prev_results,
            "context": {
                "history": history_context,
                "corrections": [
                    (qa.question, scrub_secrets.scrub_line("~/.bash_history", qa.corrected)) for qa in corrected_results
                ],
                "similar": similar_history,
            },
            "trace_id": trace_id,
        },
        headers={
            "user-agent": f"promptops-cli; user_id={user_id()}",
        },
    )
    response.raise_for_status()
    data = response.json()
    logging.debug(f"explanation response: {data}")
    return data["explanation"]


class ReturningThread(threading.Thread):
    def __init__(self, target, args=None, kwargs=None, daemon=False):
        super().__init__(target=self.run, daemon=daemon)
        self._return_value = None
        self._exception = None
        self._finished = False
        self.function = target
        self.args = args or []
        self.kwargs = kwargs or {}
        self._done_callbacks = []

    def run(self):
        try:
            self._return_value = self.function(*self.args, **self.kwargs)
        except Exception as e:
            self._exception = e
        self._finished = True
        for fn in self._done_callbacks:
            fn(self)

    def join(self, timeout=None) -> any:
        super().join(timeout)
        if self._exception is not None:
            raise self._exception
        return self._return_value

    def add_done_callback(self, fn):
        self._done_callbacks.append(fn)

    def result(self):
        if not self._finished:
            raise RuntimeError("not finished")
        if self._exception is not None:
            raise self._exception
        return self._return_value


def _loading_animation(spinner, should_continue: Callable[[], bool]):
    while should_continue():
        spinner.step()
        time.sleep(0.1)


def explain(
    *,
    result: Result,
    prev_questions: list[str],
    prev_results: list[list[str]],
    corrected_results: list[corrections.QATuple],
    history_context: list[str],
    similar_history: list[str],
):

    is_loading = True
    msg = random.choice(messages.EXPLAIN)
    spinner = Simple(f"{msg} {colorama.Style.BRIGHT}[ctrl+c]{colorama.Style.RESET_ALL} cancel")
    loading_thread = threading.Thread(
        target=_loading_animation,
        args=(
            spinner,
            lambda: is_loading,
        ),
        daemon=True,
    )
    loading_thread.start()

    t = ReturningThread(
        target=get_explanation,
        args=[],
        kwargs=dict(
            result=result,
            prev_questions=prev_questions,
            prev_results=prev_results,
            corrected_results=corrected_results,
            history_context=history_context,
            similar_history=similar_history,
        ),
        daemon=True,
    )
    t.start()

    try:
        explanation = t.join()
    except requests.HTTPError as e:
        is_loading = False
        spinner.clear()
        print_formatted_text(HTML(f"<ansired> │  oh snap, failed to fetch you the explanation: {e}</ansired>"))
        print()
        return
    finally:
        if is_loading:
            is_loading = False
            spinner.clear()

    indented = "\n".join(" │  " + line for line in explanation.splitlines())
    print_formatted_text(to_formatted_text(HTML(indented), style="fg: ansiyellow"))
    print()



import sys
from typing import Optional

from promptops import settings
import os
import uuid
from functools import lru_cache
from promptops.ui.prompts import confirm, EXIT, GO_BACK
from promptops.ui import selections
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML
import requests
from promptops.trace import trace_id
from promptops.settings_store import save, set_index_history


@lru_cache(maxsize=1)
def user_id() -> str:
    real_path = os.path.expanduser(settings.user_id_path)
    if not os.path.exists(real_path):
        uid = f"user-{uuid.uuid4().hex}"
        os.makedirs(os.path.dirname(real_path), exist_ok=True)
        with open(real_path, "w") as f:
            f.write(uid)
        return uid
    with open(real_path) as f:
        return f.read().strip()


def user_agent() -> str:
    return f"promptops-cli; user_id={user_id()}"


def has_registered() -> bool:
    real_path = os.path.expanduser(settings.user_id_path)
    return os.path.exists(real_path)


def config_flow() -> Optional[dict]:
    config_selections = {}

    print()
    print_formatted_text(
        HTML("  👋 thanks for installing <ansigreen><b>um</b></ansigreen>! let's make sure you get the most out of it")
    )
    print()

    options = ["continue", "skip"]
    ui = selections.UI(options, is_loading=False)
    selection = ui.input()
    if selection == 1:
        return

    print()
    print()
    print("  📖 indexing your history can greatly improve the speed and the quality of the suggestions")
    print_formatted_text(
        HTML(
            "  do you want to index your history? <b><i>we take special care to"
            " scrub secrets before indexing</i></b>"
        )
    )
    print()

    options = ["confirm", "skip"]
    ui = selections.UI(options, is_loading=False)
    selection = ui.input()
    if selection != 1:
        from promptops.history import index_history

        set_index_history(True)
        index_history(show_progress=True)
        config_selections["loaded_history"] = True

    print()
    while True:
        print()
        context_size = 2
        print(
            f"  ↩️  say more with less! the number of previous commands to include"
            f" in the query context (default: {context_size})?"
        )
        print()
        options = ["confirm defaults", "change"]
        ui = selections.UI(options, is_loading=False)
        selection = ui.input()
        if selection == 0:
            from promptops.settings_store import set_history_context

            set_history_context(context_size)
            config_selections["context_size"] = context_size
            break
        else:
            while True:
                choice = confirm(str(context_size)).strip()
                if choice == GO_BACK:
                    break
                if choice == EXIT:
                    return
                if choice.isdigit():
                    # TODO: validate
                    break
            if choice == GO_BACK:
                continue
            from promptops.settings_store import set_history_context

            set_history_context(int(choice))
            config_selections["context_size"] = int(choice)
            break

    print()
    print("  🎉 all done! 🎉")
    print()

    return config_selections


def register():
    print()
    print("  📢 provide your email address to receive occasional updates and tips (optional, leave blank to skip)")
    email = confirm("").strip()
    if email in [GO_BACK, EXIT]:
        email = ""
    print()

    requests.post(
        settings.endpoint + "/installed",
        json={"trace_id": trace_id, "email": email, "platform": sys.platform, "python_version": sys.version},
        headers={"user-agent": user_agent()},
    )
    config = config_flow()
    config["trace_id"] = trace_id
    requests.post(settings.endpoint + "/config", json=config, headers={"user-agent": user_agent()})
    save()



import logging
import os.path
import numpy as np
import requests

from promptops import settings
from promptops import user
from functools import lru_cache
from promptops import trace


class VectorDB(object):
    def __init__(self):
        self.vectors = np.zeros((0, 0))
        self.objects = []

    def add(self, vector, obj):
        if len(self.vectors) == 0:
            self.vectors = np.expand_dims(vector, axis=0)
        else:
            self.vectors = np.vstack((self.vectors, vector))
        self.objects.append(obj)

    def search(self, vector, k=1, min_similarity=0.8):
        # compute cosine similarity
        if len(self.vectors) == 0:
            return []
        scores = np.dot(vector, self.vectors.T).flatten()
        # rank results
        results = np.argsort(scores)[::-1]
        return [(self.objects[i], scores[i]) for i in results[:k] if scores[i] > min_similarity]

    def index(self, obj):
        return self.objects.index(obj)

    def remove(self, index):
        self.vectors = np.delete(self.vectors, index, axis=0)
        del self.objects[index]

    def save(self, path):
        dir_name = os.path.dirname(path)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        # make sure we don't corrupt the file
        tmp_path = path + ".tmp"
        with open(tmp_path, "wb") as f:
            np.savez(f, vectors=self.vectors, objects=self.objects)
        os.rename(tmp_path, path)

    def load(self, path):
        data = np.load(path, allow_pickle=True)
        self.vectors = data["vectors"]
        self.objects = data["objects"].tolist()

    def __len__(self):
        return len(self.objects)

    def __getitem__(self, index):
        return self.objects[index]

    def __iter__(self):
        return iter(self.objects)

    def __contains__(self, item):
        return item in self.objects

    def __repr__(self):
        return f"<VectorDB {len(self)} objects>"

    def __str__(self):
        return repr(self)


@lru_cache(maxsize=1_000)
def embedding(text: str) -> np.ndarray:
    resp = requests.post(
        settings.endpoint + "/embeddings",
        json={
            "text": text,
            "trace_id": trace.trace_id,
        },
        headers={
            "user-agent": f"promptops-cli; user_id={user.user_id()}",
        },
    )
    resp.raise_for_status()

    data = resp.json()
    try:
        vector = data["embeddings"]
        return np.array(vector)
    except KeyError:
        logging.debug("response: %s", data)
        raise



import base64
import json
import requests
import threading
from datetime import datetime, timedelta
from local_runner.creds import Creds


class Issuer:
    def __init__(self, creds: Creds):
        self.creds = creds
        self.token = ""
        self.token_expires = datetime.min
        self.http_client = requests.Session()
        self.auth_url = "https://authorization-tokens.global.ctrlstack.com/token"
        self.max_refresh_interval = timedelta(hours=1)
        self.mx = threading.Lock()

    def get_token(self) -> str:
        now = datetime.utcnow()
        if self.token_expires < now:
            with self.mx:
                if self.token_expires < now:
                    token = self._get_auth_token()

                    # Parse token and make sure we don't keep expiring ones for too long
                    parts = token.split(".")
                    if len(parts) != 3:
                        raise ValueError(f"JWT should be 3 dot-delimited parts, found {len(parts)}")

                    claim_str = parts[1]
                    while len(claim_str) % 4:
                        claim_str += "="

                    claim_bytes = base64.urlsafe_b64decode(claim_str)
                    claims = json.loads(claim_bytes)

                    expires = datetime.utcfromtimestamp(claims["exp"])
                    if expires > now + self.max_refresh_interval:
                        expires = now + self.max_refresh_interval

                    self.token = token
                    self.token_expires = expires

        return self.token

    def _get_auth_token(self) -> str:
        payload = {
            "clientId": self.creds.client_id,
            "clientSecret": self.creds.client_secret
        }
        headers = {"Content-Type": "application/json"}

        response = self.http_client.post(self.auth_url, json=payload, headers=headers)
        response.raise_for_status()
        jwt = response.json().get("JWT")
        if not jwt:
            raise ValueError("Error unmarshaling response body for {}: JWT not found".format(self.auth_url))

        return jwt

    def invalidate(self):
        with self.mx:
            self.token = ""
            self.token_expires = datetime.min



import re
import os
from promptops.scrub_secrets import scrub_lines


def _is_start_line(line):
    return re.match(r"^: \d+:\d+;.+", line) is not None


CMDS_TO_EXCLUDE = [
    "um",
    "qq",
    "promptops query",
]


def accept_command(cmd: str) -> bool:
    if cmd.strip() == "":
        return False
    for cmd_to_exclude in CMDS_TO_EXCLUDE:
        if cmd == cmd_to_exclude or cmd.startswith(cmd_to_exclude + " "):
            return False
    return True


def filter_commands(commands: list[str]):
    return filter(accept_command, commands)


def reverse_readline(filename, buf_size=8192):
    """A generator that returns the lines of a file in reverse order"""
    with open(filename, "rb") as fh:
        segment = None
        offset = 0
        fh.seek(0, os.SEEK_END)
        file_size = remaining_size = fh.tell()
        extra = b""
        while remaining_size > 0:
            offset = min(file_size, offset + buf_size)
            fh.seek(file_size - offset)
            # TODO: this could fail if things don't align for unicode characters
            block = fh.read(min(remaining_size, buf_size)) + extra
            remaining_size -= buf_size
            for i in range(min(8, len(block))):
                try:
                    buffer = block[i:].decode(encoding="utf-8")
                    extra = block[:i]
                    break
                except UnicodeDecodeError:
                    continue
            else:
                extra = block
                continue
            lines = buffer.split("\n")
            # The first line of the buffer is probably not a complete line so
            # we'll save it and append it to the last line of the next buffer
            # we read
            if segment is not None:
                # If the previous chunk starts right from the beginning of line
                # do not concat the segment to the last line of new chunk.
                # Instead, yield the segment first
                if buffer[-1] != "\n":
                    lines[-1] += segment
                else:
                    yield segment
            segment = lines[0]
            # iterate in reverse
            for index in range(len(lines) - 1, 0, -1):
                if lines[index]:
                    yield lines[index]
        # Don't yield None if the file was empty
        if segment is not None:
            yield segment


class Shell:
    def __init__(self, history_file):
        self.history_file = history_file

    def get_full_history(self):
        history = self._read_history_file()
        cmds = self._get_cmds_from_lines(history)
        return scrub_lines(self.history_file, list(filter_commands(cmds)))

    def _read_history_file(self):
        fname = os.path.expanduser(self.history_file)
        with open(fname, "r", encoding="utf-8", errors="ignore") as f:
            return [line.strip() for line in f if line.strip() != ""]

    def _get_cmds_from_lines(self, history):
        raise NotImplementedError()

    def get_recent_history(self, look_back: int = 10):
        raise NotImplementedError()


class NoopShell(Shell):
    def _get_cmds_from_lines(self, history):
        pass

    def __init__(self):
        super().__init__(None)

    def get_full_history(self):
        return []

    def get_recent_history(self, look_back: int = 10):
        return []



from typing import Any
from local_runner.comms import Message


def censor(value: Any) -> Any:
    if isinstance(value, dict):
        return {key: censor(val) if key.lower() != "token" else "***********" for key, val in value.items()}
    elif isinstance(value, list):
        return [censor(val) for val in value]
    return value


def censored(message: Message) -> Message:
    return Message(
        type=message.type,
        payload=censor(message.payload)
    )



import logging
import os
import subprocess
from typing import Optional

import requests
import sys

from promptops import settings
from promptops import trace
from promptops import user
from promptops.ui import selections


def init_recipe(
    q: str,
):
    req = {
        "prompt": q,
        "trace_id": trace.trace_id,
        "platform": sys.platform,
        "shell": os.environ.get("SHELL"),
    }
    response = requests.post(
        settings.endpoint + "/recipe",
        json=req,
        headers={
            "user-agent": f"promptops-cli; user_id={user.user_id()}",
        },
    )
    if response.status_code != 200:
        # this exception completely destroys the ui
        raise Exception(f"there was problem with the response, status: {response.status_code}")

    data = response.json()

    if not data.get("steps"):
        raise Exception(f"missing steps")

    return data


def run(script: str, lang: str="shell") -> (int, Optional[str]):
    if lang == "shell":
        proc = subprocess.run(
            script, shell=True, start_new_session=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        if proc.stdout and str(proc.stdout) != "":
            sys.stdout.write(proc.stdout.decode("utf-8"))
        if proc.stderr and str(proc.stderr) != "":
            sys.stdout.write(proc.stderr.decode("utf-8"))
        sys.stdout.flush()

        # history.add(scrub_secrets.scrub_line(".bash_history", cmd.script), proc.returncode)
        return proc.returncode, str(proc.stderr)
    else:
        raise NotImplementedError(f"{lang} not implemented yet")


def execute_step(step):
    print(f"executing step \n command: {step.get('command')}")
    print()

    for parameter in step.get("parameters"):
        print(f"parameter: {parameter.get('name')}")
        options = ["describe"]

        if parameter.get("resolve"):
            options.append("resolve")
        if parameter.get("create"):
            options.append("create")
        if parameter.get("options"):
            options.append("select")
        options.append("replace")
        ui = selections.UI(options, is_loading=False)
        selection = ui.input()

        option = options[selection]

        if option == "resolve" or option == "create":
            result = run(parameter.get(option))
        if option == "options":
            select_ui = selections.UI(parameter.get("options"), is_loading=False)
            selected_option = select_ui.input()
            print(selected_option)
        if options == "replace":
            value = input("enter a value for the parameter: ")



def workflow_entrypoint(prompt):
    if not prompt.question or len(prompt.question) < 1:
        print("you must include a question")
        return

    # recipe = init_recipe(prompt)

    recipe = {
        "name": "Provide a name",
        "description": "Provide a description",
        "steps": [
            {
                "command": "aws iam create-role --role-name <lambda-role-name> --assume-role-policy-document file://<path-to-trust-policy-json-file>",
                "parameters": [
                    {
                        "name": "lambda-role-name",
                        "description": "name of the Lambda role",
                        "resolve": "aws iam list-roles --query 'Roles[].RoleName' --output text"
                    },
                    {
                        "name": "path-to-trust-policy-json-file",
                        "description": "local file path for trust policy document",
                        "create": "echo '{\n            \"Version\": \"2012-10-17\",\n            \"Statement\": [\n            {\n            \"Effect\": \"Allow\",\n            \"Principal\": {\n            \"Service\": \"lambda.amazonaws.com\"\n            },\n            \"Action\": \"sts:AssumeRole\"\n            }\n            ]\n            }' > trust-policy.json"
                    }
                ]
            },
            {
                "command": "aws iam attach-role-policy --role-name <lambda-role-name> --policy-arn arn:aws:iam::aws:policy/service-role/AWSLambdaBasicExecutionRole",
                "parameters": [
                    {
                        "name": "lambda-role-name",
                        "description": "name of the Lambda execution role",
                        "resolve": "aws iam list-roles --query 'Roles[].RoleName' --output text"
                    },
                    {
                        "name": "policy-arn",
                        "description": "ARN of the AWSLambdaBasicExecutionRole policy",
                        "resolve": "aws iam list-policies --scope AWS --query 'Policies[?PolicyName==`AWSLambdaBasicExecutionRole`].Arn' --output text"
                    }
                ]
            },
            {
                "command": "aws lambda create-function --function-name <lambda-function-name> --runtime <runtime> --role <lambda-role-arn> --handler <handler> --zip-file fileb://<path-to-zip-file>",
                "parameters": [
                    {
                        "name": "<lambda-function-name>",
                        "description": "name of the Lambda function",
                        "resolve": "aws lambda list-functions --query 'Functions[].FunctionName' --output text"
                    },
                    {
                        "name": "<runtime>",
                        "description": "runtime environment for the Lambda function"
                    },
                    {
                        "name": "<lambda-role-arn>",
                        "description": "ARN of the IAM role that Lambda assumes when it executes the function",
                        "resolve": "aws iam list-roles --query 'Roles[?RoleName==`<lambda-role-name>`].Arn' --output text"
                    },
                    {
                        "name": "<handler>",
                        "description": "entry point of the Lambda function"
                    },
                    {
                        "name": "<path-to-zip-file>",
                        "description": "local file path for the ZIP file containing the Lambda function code",
                        "create": "zip -r <zip-file-name> <source-code-directory>"
                    }
                ]
            }
        ]
    }

    for step in recipe.get("steps"):
        execute_step(step)

    return



import asyncio
from local_runner import comms
from local_runner import exec
from .censor import censored
from local_runner.config import Config
from local_runner.comms import dtos
import time
import logging


class Handler:
    def __init__(self, process_mgr: exec.Manager, reporter: comms.Reporter, cfg: Config):
        self.process_mgr = process_mgr
        self.reporter = reporter
        self.cfg = cfg

    async def handle(self, data: dict, _):
        message = comms.Message.from_dict(data)
        logging.debug(f"got message: {censored(message)}")
        if message.type == "request":
            req = comms.Request.from_dict(message.payload)
            truncated = "<EMPTY>"
            if req.code.content:
                truncated = req.code.content.split("\n")[0]
                if len(truncated) > 50:
                    truncated = truncated[:50] + "..."
            logging.info(f"received {truncated}")
            exec_request = self.make_exec_request(req)
            await self.process_mgr.queue(exec_request)

    def make_exec_request(self, request: comms.Request) -> exec.Request:
        def on_started(stdout, stderr):
            print(f"started session: {request.session_id}")

            async def report(stream: str, data: str):
                payload = dtos.StreamData(stream=stream, data=data)
                self.reporter.report_update(request, dtos.UpdateMessage(
                    message_type="stream",
                    data=payload,
                    block_index=request.block_index,
                    time=int(time.time() * 1000)
                ))

            async def read_stream(stream_name: str, stream):
                async for line in stream:
                    # line is bytes
                    decoded = line.decode("utf-8")
                    print(f"{stream_name}> {decoded.rstrip()}")
                    await report(stream_name, decoded)

            asyncio.create_task(read_stream("stdout", stdout))
            asyncio.create_task(read_stream("stderr", stderr))

        def on_completed(result: exec.Result):
            print(f"Process completed, session: {request.session_id}, result: {result}")
            api_result = dtos.Result(
                status=result.exit_code if result.exit_code is not None else 1,
                reason=str(result.error) if result.error else None,
                result={},
                block_index=request.block_index,
            )
            if result.error and api_result.status == 0:
                api_result.status = 1
            self.reporter.report_result(request, api_result)

        if request.code.type == "python":
            return exec.Request(cmd=self.cfg.python_path, args=["-c", request.code.content], on_started=on_started,
                                on_completed=on_completed)
        elif request.code.type == "shell":
            return exec.Request(cmd=self.cfg.shell_path, args=["-c", request.code.content], on_started=on_started,
                                on_completed=on_completed)
        raise ValueError(f"Unknown code type: {request.code.type}")



import sys
import colorama

LINE_UP = '\033[1A'


class LoadingBase(object):
    def __init__(self):
        self._stream = sys.stdout
        self._step = 0
        self._written_lines = 0

    def step(self):
        self.clear()
        self._draw()
        self._step += 1

    def _draw(self):
        raise NotImplementedError()

    def reset(self):
        self._step = 0

    def clear(self):
        if self._written_lines == 0:
            return
        self._stream.write(colorama.ansi.clear_line() + "\r")
        for i in range(self._written_lines - 1):
            self._stream.write(LINE_UP + colorama.ansi.clear_line() + "\r")
        self._written_lines = 0



import os.path
import logging
from typing import Optional, List
import requests
import numpy as np

from promptops.similarity import VectorDB
from promptops.shells import get_shell
from promptops import settings
from promptops.user import user_id

_hist_db: Optional[VectorDB] = None


def get_history_db() -> VectorDB:
    """
    The history db records are dictionaries with the following keys:
        cmd: the command
        ignore: whether the command should be ignored
        error: whether the command is known to error
    :return:
    """
    global _hist_db
    if _hist_db is None:
        _hist_db = VectorDB()
        path = os.path.expanduser(settings.history_db_path)
        if os.path.exists(path):
            logging.debug("loading history db")
            _hist_db.load(path)
        logging.debug(f"loaded history db: {len(_hist_db)} records")
    return _hist_db


def check_history(embedding):
    history_db = get_history_db()
    results = history_db.search(embedding, k=3, min_similarity=0.8)
    results = [(obj, score) for obj, score in results if not isinstance(obj, dict) or not obj.get("ignore", False)]
    return results


def embedding_batch(cmds: List[str]) -> List[tuple[str, np.ndarray]]:
    resp = requests.post(
        settings.endpoint + "/embeddings",
        json={
            "batch": cmds,
        },
        headers={
            "user-agent": f"promptops-cli; user_id={user_id()}",
        },
    )

    try:
        resp.raise_for_status()
    except Exception as e:
        logging.debug(f"error getting embeddings: {e}, {resp.text}")
        raise
    # best effort for now
    result = resp.json().get("result", [])
    items = []
    for item in result:
        cmd = item["text"]
        vector = np.array(item["embeddings"])
        items.append((cmd, vector))

    return items


def index_history(show_progress: bool = None):
    progress = None
    if show_progress:
        from promptops.loading.progress import ProgressSpinner

        progress = ProgressSpinner(100)
        progress.increment(1)

    db = get_history_db()
    prev_commands = get_shell().get_full_history()

    if show_progress:
        progress.increment(3)

    indexed_commands = {obj if isinstance(obj, str) else obj["cmd"] for obj in db.objects}
    delta = list(set(prev_commands) - indexed_commands)
    batch_size = 32

    if show_progress is None and len(delta) > batch_size:
        from promptops.loading.progress import ProgressSpinner
        show_progress = True
        progress = ProgressSpinner(100)
        progress.increment(4)
    if show_progress:
        progress.increment(2)

    if len(delta) == 0:
        if show_progress:
            progress.set(100)
        return

    progress_inc = (100 - 6) // (len(delta) / batch_size)

    for i in range(0, len(delta), batch_size):
        for cmd, vector in embedding_batch(delta[i : i + batch_size]):
            db.add(vector, {"cmd": cmd, "ignore": False})
        if show_progress:
            progress.increment(progress_inc)

    db.save(os.path.expanduser(settings.history_db_path))

    if show_progress:
        progress.set(100)


def update_history():
    if settings.index_history:
        index_history()


def add(cmd: str, return_code: int):
    db = get_history_db()

    for obj in db.objects:
        if isinstance(obj, dict) and obj.get("cmd", None) == cmd:
            # we just need to update here
            obj["return_code"] = return_code
            db.save(os.path.expanduser(settings.history_db_path))
            return

    eb = embedding_batch([cmd])
    if len(eb) < 1:
        return
    cmd, vector = eb[0]
    db.add(vector, {"cmd": cmd, "ignore": False, "return_code": return_code})
    db.save(os.path.expanduser(settings.history_db_path))


if __name__ == "__main__":
    index_history(True)



import json
import os

from promptops import settings


config_file_path = "~/.promptops/cli-config.json"


_loaded_data = None


def load():
    real_path = os.path.expanduser(config_file_path)
    if not os.path.exists(real_path):
        return
    with open(real_path) as f:
        data = json.load(f)
    global _loaded_data
    if _loaded_data is None:
        _loaded_data = data
    settings.endpoint = data.get("endpoint", settings.endpoint)
    settings.request_explanation = data.get("request_explanation", settings.request_explanation)
    settings.model = data.get("model", settings.model)
    settings.history_context = data.get("history_context", settings.history_context)
    settings.corrections_db_path = data.get("corrections_db_path", settings.corrections_db_path)
    settings.history_db_path = data.get("history_db_path", settings.history_db_path)
    settings.index_history = data.get("index_history", settings.index_history)


def _build_data():
    data = {
        "request_explanation": settings.request_explanation,
        "model": settings.model,
        "history_context": settings.history_context,
        "corrections_db_path": settings.corrections_db_path,
        "history_db_path": settings.history_db_path,
        "index_history": settings.index_history,
    }
    if settings.endpoint != settings.DEFAULT_ENDPOINT:
        data["endpoint"] = settings.endpoint
    return data


def save():
    real_path = os.path.expanduser(config_file_path)
    os.makedirs(os.path.dirname(real_path), exist_ok=True)
    data = _build_data()
    global _loaded_data
    _loaded_data = data
    with open(real_path, "w") as f:
        json.dump(data, f, indent=2)


def set_history_context(history_context: int):
    settings.history_context = history_context
    save()


def set_index_history(index: bool):
    settings.index_history = index
    save()


def dict_diff_keys(dict1, dict2):
    keys1 = set(dict1.keys())
    keys2 = set(dict2.keys())

    diff_keys = keys1 ^ keys2
    common_keys = keys1 & keys2
    return diff_keys | {key for key in common_keys if dict1[key] != dict2[key]}


def is_changed() -> bool:
    data = _build_data()
    diff_keys = dict_diff_keys(data, _loaded_data or {})
    diff_keys.discard("endpoint")
    return len(diff_keys) > 0



import json
import os
from dataclasses import dataclass
from typing import Optional
from local_runner.config import ENV_PREFIX


@dataclass
class Creds:
    client_id: str
    client_secret: str

    def to_dict(self):
        return {
            "clientId": self.client_id,
            "clientSecret": self.client_secret
        }


class NotFoundError(Exception):
    pass


ENV_CLIENT_ID = f"{ENV_PREFIX}_CLIENT_ID"
ENV_CLIENT_SECRET = f"{ENV_PREFIX}_CLIENT_SECRET"


def default(location: str) -> Creds:
    try:
        return from_file(location)
    except FileNotFoundError:
        pass

    creds = from_env()
    if creds is not None:
        return creds

    raise NotFoundError("credentials not found")


def from_file(location: str) -> Creds:
    location = os.path.expanduser(location)
    with open(location) as f:
        creds = json.load(f)

    return Creds(client_id=creds["clientId"], client_secret=creds["clientSecret"])


def from_env() -> Optional[Creds]:
    client_id = os.getenv(ENV_CLIENT_ID)
    client_secret = os.getenv(ENV_CLIENT_SECRET)

    if client_id and client_secret:
        return Creds(client_id=client_id, client_secret=client_secret)
    else:
        return None


def store(location: str, creds: Creds) -> None:
    location = os.path.expanduser(location)
    os.makedirs(os.path.dirname(location), mode=0o700, exist_ok=True)

    with open(location, "w") as f:
        json.dump(creds.to_dict(), f, indent=2)



import colorama

from .base import LoadingBase


_chars = "⬢⬡"

alternating = "⬢⬡"
filled = "⬢"
outline = "⬡"


class PromptOps(LoadingBase):
    def __init__(self, text: str):
        super().__init__()
        self._loader = loader(text)

    def _draw(self):
        txt = self._loader() + " "
        self._stream.write(txt)
        self._stream.flush()
        self._written_lines = len(txt.split("\n"))


def loader(text: str, style: str = alternating, colors: list[str] = None):
    if colors is None:
        colors = [
            colorama.Fore.GREEN,
            colorama.Fore.CYAN,
            colorama.Fore.BLUE,
            colorama.Fore.LIGHTBLUE_EX,
            colorama.Fore.LIGHTCYAN_EX,
            colorama.Fore.LIGHTGREEN_EX,
        ]
    step = 0

    def _get_text(advance=True):
        nonlocal step
        char = style[step % len(style)]
        color = colors[step % len(colors)]
        if advance:
            step += 1
        return colorama.Style.BRIGHT + color + char + colorama.Style.RESET_ALL + "  " + text

    return _get_text



import asyncio
import random
import string
import websockets
from local_runner import creds
import json


class RegistrationHandler:
    def __init__(self, url: str, rotation_duration: float):
        self.url = url
        self.rotation_duration = rotation_duration

    async def run(self) -> creds.Creds:
        async def handle_connection(conn: websockets.WebSocketClientProtocol) -> creds.Creds:
            challenge = generate_challenge(8, 4)
            print("Code", challenge)
            await conn.send(json.dumps({
                "__action__": "register",
                "challenge": challenge,
            }))
            while True:
                response = await conn.recv()

                data = json.loads(response)
                action = data.get("__action__")

                if action == "client_credentials":
                    client_id = data.get("__client_id__")
                    client_secret = data.get("__client_secret__")
                    if client_id and client_secret:
                        return creds.Creds(client_id=client_id, client_secret=client_secret)
                    else:
                        raise ValueError("Missing client_id or client_secret in the response")
                elif action == "registration_success":
                    print("ready to pair")
                else:
                    print("Unknown action", action)

        while True:
            try:
                async with websockets.connect(self.url) as websocket:
                    print("Connected to websocket")
                    while True:
                        try:
                            c = await asyncio.wait_for(handle_connection(websocket), timeout=self.rotation_duration)
                            break
                        except asyncio.TimeoutError:
                            continue
                    print("Paired successfully")
                    return c
            except Exception as e:
                print(f"Error: {e}")
                await asyncio.sleep(5)


def generate_challenge(chars: int, group_len: int) -> str:
    allowed_chars = string.ascii_uppercase + string.digits
    result = []
    for i in range(chars):
        result.append(random.choice(allowed_chars))
        if (i + 1) % group_len == 0 and i != chars - 1:
            result.append('-')
    return ''.join(result)


async def register(url: str, rotation_duration: float) -> creds.Creds:
    return await RegistrationHandler(url, rotation_duration).run()



from pathlib import Path
from setuptools import setup, find_packages
import os


def get_long_description():
    # fmt: off
    """#PromptOps
Your CLI assistant. Ask questions, get shell commands.
"""
    # fmt: on


def get_version():
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "promptops", "version.py")
    g = {}
    with open(path) as fp:
        exec(fp.read(), g)
    return g["__version__"]


setup(
    name="promptops",
    version=get_version(),
    description="Your CLI assistant. Ask questions, get shell commands.",
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
    author="CtrlStack inc.",
    url="https://promptops.com/",
    project_urls={
        "Documentation": "https://docs.promptops.com/cli",
        "Changelog": "https://docs.promptops.com/en/stable/changelog.html",
        "Live demo": "https://promptops.com/",
    },
    license='GPLv3',
    packages=find_packages(exclude=("tests", "tests.*")),
    include_package_data=False,
    python_requires=">=3.9",
    install_requires=[
        "colorama~=0.4.6",
        "urllib3>=1.26,<2",  # kubernetes uses google-auth which has urllib3<2
        "requests~=2.29.0",
        "websockets~=11.0.2",
        "detect-secrets~=1.4.0",
        "prompt-toolkit~=3.0.38",
        "numpy~=1.24.3",
        "pyperclip~=1.8.2",
        "psutil~=5.9.5",
        "boto3~=1.26.131",
        "kubernetes~=26.1.0",
        "setuptools",
        "pip",
    ],
    entry_points="""
        [console_scripts]
        promptops=promptops.main:entry_main
        um=promptops.main:entry_alias
        qq=promptops.main:entry_alias
    """,
    setup_requires=[],
    extras_require={},
    tests_require=[],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
    ],
)



import threading
import sys
import queue


class ProgressSpinner:
    def __init__(self, total):
        self.total = total
        self.queue = queue.Queue()
        self.current = 0
        self.thread = threading.Thread(target=self.__progress_spinner, args=(self.total, self.queue))
        self.thread.start()

    def increment(self, completed):
        self.current += completed
        self.queue.put(self.current)
        if self.current >= self.total:
            self.thread.join()

    def set(self, completed):
        self.queue.put(completed)
        self.current = completed
        if completed >= self.total:
            self.thread.join()

    @staticmethod
    def __progress_spinner(total, completed_queue):
        completed = min(completed_queue.get(), total)
        percent = completed / total
        sys.stdout.write("\rProgress: [{0:50s}] {1:.1f}%".format("#" * int(percent * 50), percent * 100))
        sys.stdout.flush()

        while True:
            completed = min(completed_queue.get(), total)
            percent = completed / total
            sys.stdout.write("\rProgress: [{0:50s}] {1:.1f}%".format("#" * int(percent * 50), percent * 100))
            sys.stdout.flush()

            if completed >= total:
                sys.stdout.write("\n")
                sys.stdout.flush()
                break




from promptops.loading.base import LoadingBase
import colorama


TWO_DOTS = "⠃⠆⡄⣀⢠⠰⠘⠉"
THREE_DOTS = "⠇⠦⠴⠸⠙⠋"
FOUR_DOTS = "⡇⣆⣤⣰⢸⠹⠛⠏"


class Simple(LoadingBase):
    def __init__(self, text: str, style=THREE_DOTS):
        super().__init__()
        self._style = style
        self._text = text

    def _get_text(self):
        char = self._style[self._step % len(self._style)]
        return colorama.Fore.GREEN + char + colorama.Style.RESET_ALL + " " + self._text

    def _draw(self):
        txt = self._get_text() + " "
        self._stream.write(txt)
        self._stream.flush()
        self._written_lines = len(txt.split("\n"))


def loader(text: str, style: str = THREE_DOTS, color=colorama.Fore.GREEN):
    step = 0

    def _get_text(advance=True):
        nonlocal step
        char = style[step % len(style)]
        if advance:
            step += 1
        return colorama.Style.BRIGHT + color + char + colorama.Style.RESET_ALL + "  " + text

    return _get_text
